import threading

import click


class Spinner(threading.Thread):

    def __init__(self, msg: str, interval=0.25):
        super().__init__()
        click.secho(f"{msg} ", nl=False, fg='bright_cyan')
        self.status = threading.Event()
        self.interval = interval
        self.daemon = True

    def stop(self):
        self.status.set()

    def is_stopped(self):
        return self.status.isSet()

    def run(self):
        while not self.is_stopped():
            click.secho('.', nl=False, fg='bright_cyan')
            self.status.wait(self.interval)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        click.secho('\b', fg='bright_cyan')



import shutil
import subprocess
import sys
from typing import Optional

import click

from pip_inside.utils.pyproject import PyProject


def handle_add(name: str, group: Optional[str]):
    try:
        pyproject = PyProject.from_toml()
        name = name.lower().replace('_', '-')
        if pyproject.find_dependency(name, 'main'):
            click.secho("Skip, already installed as main dependency")
            return
        name_installed = pyproject.find_dependency(name, group)
        if name_installed:
            if name_installed == name:
                click.secho("Skip, already installed")
            else:
                pyproject.remove_dependency(name_installed, group)
        cmd = [shutil.which('python'), '-m', 'pip', 'install', name]
        if subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout).returncode == 0:
            pyproject.add_dependency(name, group)
            pyproject.flush()
    except subprocess.CalledProcessError:
        pass



import itertools
import os
from types import SimpleNamespace
from typing import List, Union

import tomlkit
from packaging.requirements import Requirement

from .version_specifies import get_package_name


class PyProject:
    def __init__(self, path='pyproject.toml') -> None:
        self.path = path
        self._meta = {}

    @classmethod
    def from_toml(cls, path='pyproject.toml'):
        pyproject = cls(path)
        pyproject.load()
        return pyproject

    def load(self):
        if not os.path.exists(self.path):
            raise ValueError(f"'{self.path}' not found")

        with open(self.path, 'r') as f:
            self._meta = tomlkit.load(f)

    def flush(self):
        with open(self.path, "w") as f:
            tomlkit.dump(self._meta, f)

    def update(self, key: str, value: Union[str, int, float, dict, list]):
        data = self._meta
        attrs = key.split('.')
        for attr in attrs[:-1]:
            data = data.setdefault(attr, {})
        data[attrs[-1]] = value

    def get(self, key: str, *, create_if_missing: bool = False, default = None):
        data = self._meta
        attrs = key.split('.')

        for attr in attrs[:-1]:
            if create_if_missing:
                data = data.setdefault(attr, {})
            else:
                data = data.get(attr)
                if data is None:
                    return default
        return data.setdefault(attrs[-1], default) if create_if_missing else data.get(attrs[-1], default)

    def set(self, key: str, value: Union[str, int, float, dict, list], *, create_if_missing: bool = True):
        data = self._meta
        attrs = key.split('.')

        for attr in attrs[:-1]:
            if create_if_missing:
                data = data.setdefault(attr, {})
            else:
                data = data.get(attr)
                if data is None:
                    return False
        data[attrs[-1]] = value
        return True

    def add_dependency(self, name: str, group: str = 'main'):
        if group == 'main':
            key = 'project.dependencies'
        else:
            key = f"project.optional-dependencies.{group}"
        dependencies = self.get(key, create_if_missing=True, default=[])
        if name not in dependencies:
            dependencies.append(name)

    def remove_dependency(self, name: str, group: str = 'main'):
        if group == 'main':
            key = 'project.dependencies'
        else:
            key = f"project.optional-dependencies.{group}"
        dependencies = self.get(key, create_if_missing=False)
        if dependencies is None or len(dependencies) == 0:
            return False
        package_name = get_package_name(name)
        remove_list = [dep for dep in dependencies if get_package_name(dep) == package_name]
        if len(remove_list) == 0:
            return False
        for dep in remove_list:
            try:
                dependencies.remove(dep)
            except ValueError:
                pass
        return True

    def find_dependency(self, name: str, group: str = 'main'):
        package_name = get_package_name(name)
        for dep in self.get_dependencies(group):
            pkg_name = get_package_name(dep)
            if pkg_name == package_name:
                return dep
        return None

    @staticmethod
    def _is_in_dependencies(name: str, dependencies: List[str]) -> bool:
        if name in dependencies:
            return True
        if name in set([get_package_name(dep) for dep in dependencies]):
            return True
        return False

    def get_dependencies(self, group: str = 'main'):
        if group == 'all':
            key_main = 'project.dependencies'
            key_optionals = 'project.optional-dependencies'
            deps_main = self.get(key_main, default=[])
            deps_optionals = list(itertools.chain(*self.get(key_optionals, default={}).values()))
            return deps_main + deps_optionals

        if group == 'main':
            return self.get('project.dependencies', default=[])
        else:
            return self.get(f"project.optional-dependencies.{group}", default=[])

    def get_dependencies_with_group(self):
        dependencies = {}
        for dep in self.get('project.dependencies', default=[]):
            dependencies[Requirement(dep)] = 'main'

        for group, deps in self.get('project.optional-dependencies', default={}).items():
            for dep in deps:
                dependencies[Requirement(dep)] = group
        return dependencies

    @staticmethod
    def get_template():
        return SimpleNamespace(
            name=os.path.basename(os.getcwd()),
            description=''
        )



import os
import shutil
import signal
import sys
import venv
from pathlib import Path

import pexpect

from pip_inside import Aborted


def handle_shell():
    if os.name != "posix":
        raise Aborted(f"Sorry, only supports *nix, : {os.name}")

    _create_venv()
    _spaw_new_shell()


def _create_venv():
    if os.path.exists('.venv'):
        return
    name = Path(os.getcwd()).name
    venv.create('.venv', with_pip=True, prompt=name)


def _spaw_new_shell():
    def resize(*args, **kwargs) -> None:
        terminal = shutil.get_terminal_size()
        p.setwinsize(terminal.lines, terminal.columns)

    shell = os.environ.get("SHELL")
    terminal = shutil.get_terminal_size()
    p = pexpect.spawn(shell, ['-i'], dimensions=(terminal.lines, terminal.columns))
    if shell.endswith('/zsh'):
        p.setecho(False)
    p.sendline('source .venv/bin/activate')
    signal.signal(signal.SIGWINCH, resize)
    p.interact(escape_character=None)
    p.close()
    sys.exit(p.exitstatus)



import hashlib
import io
import os
import tarfile
import tempfile
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from types import SimpleNamespace
from typing import Optional

import click
from flit_core.common import Metadata, Module, get_info_from_module
from flit_core.config import pep621_allowed_fields, read_flit_config
from flit_core.sdist import SdistBuilder
from flit_core.wheel import WheelBuilder

from pip_inside.utils.pyproject import PyProject


def make_metadata(module, ini_info, extra_meta):
    metadata = {'name': module.name, 'provides': [module.name]}
    metadata.update(get_info_from_module(module, ini_info.dynamic_metadata))
    metadata.update(ini_info.metadata)
    if extra_meta:
        metadata.update(extra_meta)
    return Metadata(metadata)


class PiWheelBuilder(WheelBuilder):

    @classmethod
    def build_wheel(cls, config_path: Path, dist_path: Path, extra_meta: Optional[dict] = None):
        def create_builder(target_fp):
            directory = config_path.parent
            ini_info = read_flit_config(config_path)
            entrypoints = ini_info.entrypoints
            module = Module(ini_info.module, directory)
            metadata = make_metadata(module, ini_info, extra_meta)
            return cls(
                directory, module, metadata, entrypoints, target_fp, ini_info.data_directory
            )
        (fd, temp_path) = tempfile.mkstemp(suffix='.whl', dir=str(dist_path))
        try:
            with io.open(fd, 'w+b') as fp:
                wb = create_builder(fp)
                wb.build(False)

            wheel_path = dist_path / wb.wheel_filename
            os.replace(temp_path, str(wheel_path))
        except:
            os.unlink(temp_path)
            raise

        return SimpleNamespace(builder=wb, file=wheel_path)


class PiSdistBuilder(SdistBuilder):

    @classmethod
    def build_sdist(cls, config_path: Path, dist_path: Path, extra_meta: Optional[dict] = None):

        def create_builder():
            ini_info = read_flit_config(config_path)
            srcdir = config_path.parent
            module = Module(ini_info.module, srcdir)
            metadata = make_metadata(module, ini_info, extra_meta)
            extra_files = [config_path.name] + ini_info.referenced_files
            return cls(
                module, metadata, srcdir, ini_info.reqs_by_extra,
                ini_info.entrypoints, extra_files, ini_info.data_directory,
                ini_info.sdist_include_patterns, ini_info.sdist_exclude_patterns,
            )

        builder = create_builder()
        sdist_file = builder.build(dist_path, gen_setup_py=False)
        return SimpleNamespace(builder=builder, file=sdist_file)


def handle_build(dist: str = 'dist'):
    click.secho(f"Building wheel and sdist to: {dist}", fg='bright_cyan')
    pkg = build_package(dist)

    wheel_name, sdist_name = str(pkg.wheel.file), str(pkg.sdist.file)
    pad_size = max(len(wheel_name), len(sdist_name)) + 1
    click.secho(f"Build {wheel_name: <{pad_size}} md5: {pkg.wheel_md5}, size: {pkg.wheel_size}", fg='green')
    click.secho(f"Build {sdist_name: <{pad_size}} md5: {pkg.sdist_md5}, size: {pkg.sdist_size}", fg='green')


def build_package(path_dist: str):
    os.makedirs('dist', exist_ok=True)
    pyproject = PyProject.from_toml()
    path_toml, path_dist = Path('pyproject.toml'), Path(path_dist)
    extra_meta = {
        'author': ','.join([author.get('name') for author in pyproject.get('project.authors')]),
        'license': pyproject.get('project.license-expression'),
        'home_page': pyproject.get('project.urls.homepage'),
    }

    pep621_allowed_fields.add('license-expression')
    wheel_info = PiWheelBuilder.build_wheel(path_toml, path_dist, extra_meta)
    sdist_info = PiSdistBuilder.build_sdist(path_toml, path_dist, extra_meta)

    wheel_data, sdist_data = open(wheel_info.file, 'rb').read(), open(sdist_info.file, 'rb').read()
    wheel_md5 = hashlib.md5(wheel_data).hexdigest()
    wheel_sha256 = hashlib.sha256(wheel_data).hexdigest()
    wheel_size = get_file_size(wheel_info.file)
    sdist_md5 = hashlib.md5(sdist_data).hexdigest()
    sdist_sha256 = hashlib.sha256(sdist_data).hexdigest()
    sdist_size = get_file_size(sdist_info.file)

    return SimpleNamespace(
        wheel=wheel_info,
        wheel_md5=wheel_md5,
        wheel_sha256=wheel_sha256,
        wheel_size=wheel_size,
        sdist=sdist_info,
        sdist_md5=sdist_md5,
        sdist_sha256=sdist_sha256,
        sdist_size=sdist_size,
    )


@contextmanager
def unpacked_tarball(path):
    tf = tarfile.open(str(path))
    with TemporaryDirectory() as tmpdir:
        tf.extractall(tmpdir)
        files = os.listdir(tmpdir)
        assert len(files) == 1, files
        yield os.path.join(tmpdir, files[0])


def get_file_size(filename: str) -> str:
    file_size = os.path.getsize(filename) / 1024
    size_unit = "KB"

    if file_size > 1024:
        file_size = file_size / 1024
        size_unit = "MB"

    return f"{file_size:.1f} {size_unit}"



import collections

import click
import tomlkit

from pip_inside.utils.dependencies import Package


def handle_freeze():
    pkg = Package.from_project()
    dependencies = collections.defaultdict(list)
    for child in pkg.children:
        dependencies[child.group].append(f"{child.name}=={child.version}")
        for group, dep in _find_installed_child(child):
            dependencies[group or child.group].append(dep)
    with open('pi.lock', 'w') as f:
        tomlkit.dump(dependencies, f)
    click.secho(f"Generated pi.lock", fg='bright_cyan')


def _find_installed_child(pkg: Package):
    for child in pkg.children:
        yield child.group, f"{child.name}=={child.version}"
        for group, dep in _find_installed_child(child):
            yield group or child.group, dep



import configparser
import os
from types import SimpleNamespace

import click
import requests
from flit_core.common import Metadata
from InquirerPy import inquirer

from pip_inside import Aborted
from pip_inside.utils import spinner

from .build import build_package

API_PYPI = 'https://upload.pypi.org/legacy/'
API_TESTPYPI = 'https://test.pypi.org/legacy/'
WEB_PYPI = "https://pypi.org/"
WEB_TESTPYPI = "https://test.pypi.org/"
NAME_TO_API = {
    '': API_PYPI,
    None: API_PYPI,
    'pypi': API_PYPI,
    'testpypi': API_TESTPYPI,
}
API_NORMALIZE = {
    "http://pypi.python.org/": API_PYPI,
    "https://pypi.python.org/": API_PYPI,
    "http://test.pypi.org/": API_TESTPYPI,
    "https://test.pypi.org/": API_TESTPYPI,
    "http://testpypi.python.org/": API_TESTPYPI,
    "https://testpypi.python.org/": API_TESTPYPI,
}
API_TO_WEB = {
    API_PYPI: WEB_PYPI,
    API_TESTPYPI: WEB_TESTPYPI
}


def handle_publish(
    repository: str = 'pypi',
    *,
    dist: str = 'dist',
    config_file: str = '~/.pypirc',
    interactive: bool = False
):
    credential = get_credential_from_pypirc(repository, config_file, interactive)
    pkg = build_package(dist)
    upload_to_repository(pkg, credential)


def get_credential_from_pypirc(repository: str, config_file: str = '~/.pypirc', interactive: bool = False):
    cp = configparser.ConfigParser()
    config_file = os.path.expanduser(config_file)
    cp.read(config_file)

    url = cp.get(repository, 'repository', fallback=None)
    url = NAME_TO_API.get(repository) if url is None else API_NORMALIZE.get(url, url)
    if url is None:
        raise Aborted(f"Missing key: 'repository' in '{repository}' section of '{config_file}'")

    if url.startswith('http://'):
        click.secho(f"Warning: insecure repository: {url}, has risk of credential leaking risk", fg='yellow')

    username = cp.get(repository, 'username', fallback=None)
    password = cp.get(repository, 'password', fallback=None)

    credentials = SimpleNamespace(name=repository, url=url, username=username, password=password)
    check_credentials(credentials, config_file, interactive)
    return credentials


def check_credentials(credential, config_file: str, interactive: bool):
    if credential.username is None:
        if not interactive:
            raise Aborted(f"'username' expected in {config_file}")
        credential.username = inquirer.text(message='Username:', mandatory=True).execute().strip()

    if credential.password is None:
        if not interactive:
            raise Aborted(f"'password' expected in {config_file}")
        credential.username = inquirer.text(message='Password:', mandatory=True).execute().strip()


def upload_to_repository(pkg, credential):
    def upload(metadata: Metadata, file, md5_digest, sha256_digest):
        with spinner.Spinner(f"Uploading {file.name}"):
            data = build_post_data(metadata, file.suffix, md5_digest, sha256_digest)
            with file.open('rb') as f:
                content = f.read()
                files = {'content': (file.name, content)}
            resp = requests.post(credential.url, data=data, files=files, auth=(credential.username, credential.password))
            resp.raise_for_status()
    try:
        click.secho(f"Publishing to [{credential.name}] ({credential.url})", fg='bright_cyan')
        upload(pkg.wheel.builder.metadata, pkg.wheel.file, pkg.wheel_md5, pkg.wheel_sha256)
        upload(pkg.sdist.builder.metadata, pkg.sdist.file, pkg.sdist_md5, pkg.sdist_sha256)
        check_published_url(credential.url, pkg.wheel.builder.metadata)
    except requests.exceptions.HTTPError as e:
        import traceback
        click.secho(traceback.format_exc(), fg='red')
        raise Aborted(f"Failed to upload, {e}")


def build_post_data(metadata: Metadata, ext: str, md5_digest, sha256_digest):
    params_of_ext = {
        '.whl': {'filetype': 'bdist_wheel', 'pyversion': 'py3'},
        '.gz': {'filetype': 'sdist'}
    }
    params_general = {
        ':action': 'file_upload',
        'name': metadata.name,
        'version': metadata.version,
        'metadata_version': '2.1',
        'summary': metadata.summary,
        'home_page': metadata.home_page,
        'author': metadata.author,
        'author_email': metadata.author_email,
        'maintainer': metadata.maintainer,
        'maintainer_email': metadata.maintainer_email,
        'license': metadata.license,
        'description': metadata.description,
        'keywords': metadata.keywords,
        'platform': metadata.platform,
        'classifiers': metadata.classifiers,
        'download_url': metadata.download_url,
        'supported_platform': metadata.supported_platform,
        # Metadata 1.1 (PEP 314)
        'provides': metadata.provides,
        'requires': metadata.requires,
        'obsoletes': metadata.obsoletes,
        # Metadata 1.2 (PEP 345)
        'project_urls': metadata.project_urls,
        'provides_dist': metadata.provides_dist,
        'obsoletes_dist': metadata.obsoletes_dist,
        'requires_dist': metadata.requires_dist,
        'requires_external': metadata.requires_external,
        'requires_python': metadata.requires_python,
        # Metadata 2.1 (PEP 566)
        'description_content_type': metadata.description_content_type,
        'provides_extra': metadata.provides_extra,

        'protocol_version': 1,
        'md5_digest': md5_digest,
        'sha256_digest': sha256_digest,
    }
    params = {**params_of_ext.get(ext), **params_general}
    return {k: v for k, v in params.items() if v}


def check_published_url(url, meta: Metadata):
    pypi_web = API_TO_WEB.get(url)
    if pypi_web:
        published_url = f"{pypi_web}project/{meta.name}/{meta.version}"
        click.secho(f"View at: {published_url}", fg='bright_cyan')
    click.secho('Done', fg='green')



import shutil
import subprocess
import sys

import click

from pip_inside.utils.dependencies import Package
from pip_inside.utils.pyproject import PyProject


def handle_remove(name, group):
    try:
        pyproject = PyProject.from_toml()
        if pyproject.remove_dependency(name, group):
            pyproject.flush()
            deps = Package.get_unused_sub_dependencies(name)
            if deps:
                cmd = [shutil.which('python'), '-m', 'pip', 'uninstall', name, *deps, '-y']
            else:
                cmd = [shutil.which('python'), '-m', 'pip', 'uninstall', name, '-y']
            subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)
        else:
            click.secho(f"Package: [{name}] not found in group: [{group}]", fg='yellow')
    except subprocess.CalledProcessError:
        pass



import re

P_VERSION_HAS_SPECIFIERS = re.compile('^[a-zA-Z0-9_.-]+\s*(?=(?:===|~=|==|!=|<=|>=|<|>))')
P_HAS_VERSION_SPECIFIERS = re.compile('(?:===|~=|==|!=|<=|>=|<|>)')
P_NORMALIZE = re.compile('^[a-zA-Z0-9_.-]+\s*(?=(?:===|~=|==|!=|<=|>=|<|>)?\s*;?)')
URL_VERSION_SPECIFIERS = 'https://peps.python.org/pep-0440/#version-specifiers'


def ver_has_spec(name: str):
    return P_VERSION_HAS_SPECIFIERS.search(name) is not None


def has_ver_spec(name: str):
    return P_HAS_VERSION_SPECIFIERS.search(name) is not None


def get_package_name(name: str):
    return P_NORMALIZE.search(norm_name(name)).group()


def norm_name(name: str):
    return name.lower().replace('_', '-') if name else None



import collections
import re
import shutil
import subprocess
from datetime import datetime
from typing import Union

import click
import requests
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from . import spinner, version_specifies

try:
    from importlib.metadata import PackageNotFoundError, distribution
except ImportError:
    from pkg_resources import DistributionNotFound as PackageNotFoundError
    from pkg_resources import get_distribution as distribution


API_URL = "https://pypi.org/search/?q={query}"
DATE_FORMAT = '%Y-%m-%d'

P_NAME = re.compile(r"<span class=\"package-snippet__name\">(.+)</span>")
P_VERSION = re.compile(r".*<span class=\"package-snippet__version\">(.+)</span>")
P_RELEASE = re.compile(r"<time\s+datetime=\"([^\"]+)\"")
P_DESCRIPTION = re.compile(r".*<p class=\"package-snippet__description\">(.+)</p>")

P_INDEX_VERSIONS = re.compile('(?<=Available versions:)([a-zA-Z0-9., ]+)')


def prompt_a_package(again: bool = False):
    prompt = 'Add aother package (leave blank to exit):' if again else 'Add a package (leave blank to exit):'
    name = inquirer.text(message=prompt).execute()
    if not name:
        return

    with spinner.Spinner(f"Searching for {name}"):
        pkgs = search(name)
    name = inquirer.select(
        message="Select the package:",
        choices=[Choice(value=pkg.name, name=pkg.desc) for pkg in pkgs],
        vi_mode=True,
        wrap_lines=True,
        mandatory=True,
    ).execute()

    with spinner.Spinner(f"Fetching version list for {name}"):
        versions = fetch_versions(name)
    if versions:
        version = inquirer.fuzzy(
            message="Select the version:",
            choices=['[set manually]'] + versions[:15],
            vi_mode=True,
            wrap_lines=True,
            mandatory=True,
        ).execute()
        if version == '[set manually]':
            version = inquirer.text(message="Version:", completer={v: None for v in versions[:15]}).execute().strip()
    else:
        click.secho('Failed to fetch version list, please set version menually', fg='cyan')
        version = inquirer.text(message="Version:").execute().strip()
    if version:
        name = f"{name}{version}" if version_specifies.has_ver_spec(version) else f"{name}=={version}"
    return name


def check_version(package_name: str) -> Union[str, bool]:
    try:
        installed = distribution(package_name)
    except PackageNotFoundError:
        return False
    else:
        return installed.version


def search(name: str):
    url = API_URL.format(query=name)
    page_data = requests.get(url=url).text
    names = P_NAME.findall(page_data)
    versions = P_VERSION.findall(page_data)
    releases = P_RELEASE.findall(page_data)
    descriptions = P_DESCRIPTION.findall(page_data)
    releases = [
        datetime.strptime(release, "%Y-%m-%dT%H:%M:%S%z").strftime(DATE_FORMAT)
        for release in releases
    ]

    n_n = max(map(len, names)) + 1
    n_v = max(map(len, versions)) + 1
    n_r = max(map(len, releases)) + 1
    n_d = max(map(len, descriptions)) + 1

    fmt = lambda n, v, r, d: f"{n: <{n_n}} {v: <{n_v}} {r: <{n_r}} {d: <{n_d}}"
    pkg = collections.namedtuple('pkg', ['name', 'desc'])

    return [
        pkg(name, fmt(name, version, release, desc))
        for name, version, release, desc in zip(names, versions, releases, descriptions)
    ]


def fetch_versions(name: str):
    return versions_by_pip_index(name) or versions_by_json(name)


def versions_by_pip_index(name: str):
    try:
        cmd = [shutil.which('python'), '-m', 'pip', 'index', 'versions', name]
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, _ = process.communicate()
        m = P_INDEX_VERSIONS.search(out.decode())
        if m is None:
            return None
        return [v.strip() for v in m.group().strip().split(',')]
    except subprocess.CalledProcessError:
        return None


def versions_by_json(name: str):
    try:
        url = f"https://pypi.org//pypi/{name}/json"
        data = requests.get(url).json()
        if data is None or len(data) < 10:
            return None
        versions = list(data.get('releases').keys())
        versions.reverse()
        return versions
    except Exception:
        return None



from importlib.util import module_from_spec, spec_from_file_location

import click

from pip_inside.utils.pyproject import PyProject


def handle_version(short: bool = False):
    pyproject = PyProject.from_toml()
    module = pyproject.get('project.name')
    s = spec_from_file_location(module, f"{module.lower().replace('-', '_')}/__init__.py")
    m = module_from_spec(s)
    s.loader.exec_module(m)
    version = m.__version__ if short else f"{module}: {m.__version__}"
    click.secho(version, fg='bright_cyan')



import itertools
import os
import shutil
import subprocess
import sys
from typing import List

import click
import tomlkit

from pip_inside.utils.pyproject import PyProject


def handle_install(groups: List[str]):
    _install_from_pi_lock(groups) or _install_from_pyproject_toml(groups)


def _install_from_pi_lock(groups: List[str]):
    if not os.path.exists('pi.lock'):
        return False
    try:
        with open('pi.lock', 'r') as f:
            data = tomlkit.load(f)

        deps = list(itertools.chain(*[data.get(group, []) for group in groups]))
        cmd = [shutil.which('python'), '-m', 'pip', 'install', *deps]
        subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)
    except subprocess.CalledProcessError:
        pass


def _install_from_pyproject_toml(groups: List[str]):
    try:
        pyproject = PyProject.from_toml()
        dependencies = []
        for group in groups:
            deps = pyproject.get_dependencies(group)
            if deps is None:
                click.secho(f"Dependencies group: {group} not found in pyproject.toml", fg='yellow')
                continue
            dependencies.extend(deps)
        if len(dependencies) == 0:
            click.secho('Nothing to install, no dependencies specified in pyproject.toml')
            return
        cmd = [shutil.which('python'), '-m', 'pip', 'install', *dependencies]
        subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)
    except subprocess.CalledProcessError:
        pass



import os
import pathlib
import subprocess
from datetime import date
from types import SimpleNamespace

import click
import tomlkit
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from pip_inside import Aborted
from pip_inside.utils import packages
from pip_inside.utils.licenses import LICENSES


def handle_init():
    click.echo(f"Initializing project in {os.getcwd()}")
    check_pyproject_toml()
    meta = collect_metadata()
    toml = build_toml(meta)

    click.secho(tomlkit.dumps(toml), fg='blue')
    proceed = inquirer.confirm(message="Proceed?", default=True).execute()
    if not proceed:
        return

    write_toml(toml)
    write_readme(meta)
    write_license(meta)
    write_root_module_with_version(meta)


def check_pyproject_toml():
    if not os.path.exists('pyproject.toml'):
        return

    try:
        with open('pyproject.toml', 'r') as f:
            data = tomlkit.load(f)
        if 'build-system' in data:
            raise Aborted(f"Skip, 'pyproject.toml' file with 'build-system' already exists")
    except Exception as e:
        raise Aborted(f"Skip, 'pyproject.toml' already exists, and failed to load: {e}")


def collect_metadata():
    defaults = get_defaults()
    name = inquirer.text(message="Name:", default=defaults.name, mandatory=True).execute()
    description = inquirer.text(message="Description:").execute()
    version = inquirer.text(message="Version:", default=defaults.version, mandatory=True).execute()
    author = inquirer.text(message="Author name:", default=defaults.author_name).execute()
    email = inquirer.text(message="Author email:", default=defaults.author_email).execute()

    license_choices = [Choice('skip')] + [Choice(value=n, name=f"{n} ({d})") for n, d in LICENSES.items()]
    license_name = inquirer.fuzzy(message="License:", choices=license_choices, vi_mode=True, wrap_lines=True).execute()
    requires_python = inquirer.text(message="requires-python:", default=defaults.requires_python, mandatory=True).execute()
    homepage = inquirer.text(message="Home page:").execute()
    dependencies = []

    if inquirer.confirm(message="Add dependencies?", default=True).execute():
        dependencies = collect_dependencies()

    return SimpleNamespace(
        name=name,
        description=description,
        version=version,
        author=author,
        email=email,
        license=license_name,
        requires_python=requires_python,
        homepage=homepage,
        dependencies=dependencies
    )


def collect_dependencies():
    dependencies = []
    name = packages.prompt_a_package()
    while name is not None:
        dependencies.append(name)
        name = packages.prompt_a_package(True)
    return dependencies


def get_defaults():
    author_name, author_email = get_default_author()
    return SimpleNamespace(
        name=os.path.basename(os.getcwd()).lower(),
        version='0.1.0',
        author_name=author_name,
        author_email=author_email,
        requires_python='>=3.8',
    )


def get_default_author():
    try:
        cmd = ['git', 'config', '-l']
        git_configs = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode()
        values = dict(s.split('=', 1) for s in git_configs.splitlines())
        return values.get('user.name'), values.get('user.email')
    except Exception:
        return None, None


def build_toml(meta):
    toml = tomlkit.document()
    project_table = tomlkit.table()
    project_table.add('name', meta.name)
    project_table.add('description', meta.description)
    if meta.author or meta.email:
        authors_list = tomlkit.array()
        author_inline = tomlkit.inline_table()
        author_inline.update({'name': meta.author or '', 'email': meta.email or ''})
        authors_list.append(author_inline)
        project_table.add('authors', authors_list)
    project_table.add('readme', 'README.md')
    license_inline = tomlkit.inline_table()
    license_inline.update({'file': 'LICENSE'})
    project_table.add('license', license_inline)
    project_table.add('license-expression', meta.license)
    project_table.add('dynamic', tomlkit.array('["version"]'))
    project_table.add('requires-python', meta.requires_python)
    dependencies_list = tomlkit.array()
    dependencies_list.extend(meta.dependencies)
    project_table.add('dependencies', dependencies_list)
    toml.add('project', project_table)

    if meta.homepage:
        urls_table = tomlkit.table()
        urls_table.add('homepage', meta.homepage)
        project_table.add('url', urls_table)

    build_system_table = tomlkit.table()
    build_system_table.add('requires', tomlkit.array('["flit_core>=3.8.0,<4"]'))
    build_system_table.add('build-backend', 'flit_core.buildapi')
    toml.add('build-system', build_system_table)

    return toml


def write_toml(toml):
    with open('pyproject.toml', 'w') as f:
        tomlkit.dump(toml, f)
        click.secho(f"Added 'pyproject.toml'", fg='bright_cyan')


def write_readme(meta):
    if os.path.exists('README.md'):
        return
    with open('README.md', 'w') as f:
        f.write(f"# {meta.name}\n\n {meta.description}\n")
        click.secho(f"Added 'README.md'", fg='bright_cyan')


def write_license(meta):
    if os.path.exists('LICENSE') or meta.license == 'skip':
        return
    licenses_dir = pathlib.Path(__file__).parent / 'licenses'
    with (licenses_dir / f"{meta.license}.txt").open() as f_in, open('LICENSE', 'w') as f_out:
        year = date.today().year
        f_out.write(f_in.read().format(year=year, author=meta.author))
        click.secho(f"Added 'LICENSE'", fg='bright_cyan')


def write_root_module_with_version(meta):
    module_name = meta.name.lower().replace('-', '_')
    os.makedirs(module_name, exist_ok=True)
    with open(f"{module_name}/__init__.py", 'w') as f:
        f.write(f"__version__ = '{meta.version}'")
        click.secho(f"Added '{module_name}/__init__.py'", fg='bright_cyan')
